import functools

import v3io.dataplane.transport


class Batch(object):

    def __init__(self, client):
        self._client = client
        self._encoded_requests = []
        self._inflight_requests = []
        self._transport_actions = v3io.dataplane.transport.Actions.encode_only
        self._transport = self._client._transport

        for client_call in [
            'get_containers',
            'get_container_contents',
            'get_object',
            'put_object',
            'delete_object',
            'put_item',
            'update_item',
            'get_item',
            'get_items',
            'create_stream',
            'delete_stream',
            'describe_stream',
            'seek_shard',
            'put_records',
            'get_records'
        ]:
            setattr(self, client_call, functools.partial(self._call_client, client_call))

    def _call_client(self, name, *args, **kw_args):
        kw_args['transport_actions'] = self._transport_actions
        request = getattr(self._client, name)(*args, **kw_args)

        self._encoded_requests.append(request)

    def wait(self):
        responses = []

        # while we can send requests - send them
        while self._encoded_requests and len(self._inflight_requests) < self._transport.max_connections:

            # send the request
            request = self._transport.send_request(self._encoded_requests.pop(0))

            # add to inflight requests
            self._inflight_requests.append(request)

        # start creating responses
        while self._inflight_requests:

            # get an inflight request
            inflight_request = self._inflight_requests.pop(0)

            # wait for the response of the request
            response = self._transport.wait_response(inflight_request)

            # add to responses
            responses.append(response)

            # if there's a pending request, send it on the connection that we just read from
            if self._encoded_requests:

                # send the request
                request = self._transport.send_request(self._encoded_requests.pop(0),
                                                       transport_state=inflight_request.transport)

                # add to inflight requests
                self._inflight_requests.append(request)

        # todo - raise for, etc
        return responses



import requests

import v3io.dataplane.response
import v3io.dataplane.request
from . import abstract


class Transport(abstract.Transport):

    def __init__(self, logger, endpoint=None, max_connections=4, timeout=None):
        super(Transport, self).__init__(logger, endpoint, max_connections, timeout)
        self._next_connection_pool = 0
        self._session = requests.Session()

    def close(self):
        self._session.close()

    def send_request(self, request, transport_state=None):

        # call the encoder to get the response
        http_response = self._http_request(request.method,
                                           request.path,
                                           request.headers,
                                           request.body)

        # set http response
        setattr(request.transport, 'http_response', http_response)

        return request

    def wait_response(self, request):

        # create a response
        return v3io.dataplane.response.Response(request.output,
                                                request.transport.http_response.status_code,
                                                request.headers,
                                                request.transport.http_response.content)

    def _http_request(self, method, path, headers=None, body=None):
        # self._logger.debug_with('Tx', method=method, path=path, headers=headers, body=body)

        response = self._session.request(method,
                                         self._endpoint + path,
                                         headers=headers,
                                         data=body,
                                         timeout=self._timeout,
                                         verify=False)

        # self._logger.debug_with('Rx', status_code=response.status_code, headers=response.headers, body=response.text)

        return response



def url_join(base, *parts):
    result = base

    if result[0] != '/':
        result = '/' + base

    for part in parts:
        if part[0] != '/':
            result += '/' + part
        else:
            result += part

    return result



import os

import v3io.dataplane.request
import v3io.dataplane.transport


class Transport(object):

    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None):
        self._logger = logger
        self._endpoint = self._get_endpoint(endpoint)
        self._timeout = timeout
        self.max_connections = max_connections or 4

    def request(self,
                container,
                access_key,
                raise_for_status,
                transport_actions,
                encoder,
                encoder_args,
                output=None):

        # default to sending/receiving
        transport_actions = transport_actions or v3io.dataplane.transport.Actions.send_and_receive

        # allocate a request
        request = v3io.dataplane.request.Request(container,
                                                 access_key,
                                                 raise_for_status,
                                                 encoder,
                                                 encoder_args,
                                                 output)

        # if all we had to do is encode, return now
        if transport_actions == v3io.dataplane.transport.Actions.encode_only:
            return request

        # send the request
        inflight_request = self.send_request(request)

        # wait for the response
        return self.wait_response(inflight_request)

    @staticmethod
    def _get_endpoint(endpoint):

        if endpoint is None:
            endpoint = os.environ.get('V3IO_API')

            if endpoint is None:
                raise RuntimeError('Endpoints must be passed to context or specified in V3IO_API')

        if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
            endpoint = 'http://' + endpoint

        return endpoint.rstrip('/')
